/************************************************************************
|                          EXAR Corporation
|                      Data Communications Division
|         DOS Test Program Source code for XR88C92/192 & XR68C92/192
|************************************************************************
|
[Date]           April 02, 2002

[Disclaimer]
 2002 EXAR Corporation
All Rights Reserved.

Exar Corporation ("Exar") hereby grants the User of this Exar product-specific
source code in "C" a non-exclusive, nontransferable license to use this Exar
product-specific source code under the following terms. Before using this Exar
product-specific source code, the User should read the following use
restrictions.  If the User does not accept these terms, the Exar
product-specific source code should not be used or downloaded.

The User is granted this license only to use the Exar product-specific source
code for User's internal use, and is not granted rights to sell, loan,
rent, lease or license the Exar product-specific source code in whole or in
part, or in modified form to anyone other than User.  User may modify the
Exar product-specific source code to suit its specific applications but
rights to derivative works and such modifications shall belong to Exar.

This Exar product-specific source code is provided on an "AS IS" basis and
Exar makes absolutely no warranty with respect to the information contained
herein.  EXAR DISCLAIMS AND USER WAIVES ALL WARRANTIES, EXPRESS OR IMPLIED,
INCLUDING WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
The entire risk as to quality and performance is with the User.
ACCORDINGLY, IN NO EVENT SHALL EXAR BE LIABLE FOR ANY DAMAGES, WHETHER IN
CONTRACT OR TORT, INCLUDING ANY LOST PROFITS OR OTHER INCIDENTAL,
CONSEQUENTIAL, EXEMPLARY, OR PUNITIVE DAMAGES ARISING OUT OF THE USE OR
APPLICATION OF THE EXAR PRODUCT-SPECIFIC SOURCE CODE.  Further, Exar reserves
the right to make changes without notice to improve reliability, function or
design.  Exar does not convey any license under patent rights or any other
intellectual property rights, including those of third parties.

Exar is not obligated to provide maintenance or support for the Exar
product-specific source code.
[Copyright]      Copyright 2002 EXAR Corporation.

*****************************************************************************
*****************************************************************************/
/*            XR88C92/192 and XR68C92/192 Test Program
Copyright - EXAR Corporation
Rev 1.1 as of Nov. 16 2001 - Code by BL
This program is best used with Exar's XR88C92/192 or XR68C92/192
Evaluation Board. The base address is automatically found out and
the IRQ used is IRQ5. Please make sure you select IRQ5 on the board. */

#include <iostream.h>
#include <conio.h>
#include <stdio.h>
#include <dos.h>
#include "8892.h"
#define intvect  0x0d		//default interrupt vector for IRQ5
#define	SIZE   	30000


 unsigned baseaddress;
 int temp;
 int fifo = 8;
 int ctsa = 0;
 int ctsb = 0;
 int error = 0;
 int imr = 0x22;
 int csra;
 int csrb;
 int acr;
 int mr0_bit0;
 char baud;
 int tx_trg = 1;
 int rx_trg = (fifo - 2);
 char loopback;
 int chipscan (void);
 int init (void);
 int send (void);
 inline int RXRDYA (void) {return (read(SRA) & 0x01);}
 inline int TXRDYA (void) {return (read(SRA) & 0x04);}
 inline int RXRDYB (void) {return (read(SRB) & 0x01);}
 inline int TXRDYB (void) {return (read(SRB) & 0x04);}

 long int bufferin0, bufferout0, bufferin1, bufferout1;
 char *buffer0, *buffer1;
 char data0 [64];
 char data1 [64];
 int databuf0 = 0;
 int databuf1 = 0;

/* oldportisr is the existing interrupt service routine */

void interrupt (*oldportisr) (...);

/* myportisr() is the new interrupt service routine
	 which will be used for this program */

void interrupt myportisr(...)
{
 int int_val = read (ISR) & imr;
 temp = int_val;
 int i;
 write (IMR, 0x00);

  if (int_val & 0x20) {
	while (RXRDYB() ) {
			buffer1[bufferin1] = read (RHRB);
			bufferin1++;
			if (bufferin1 == SIZE) bufferin1 = 0;
	}
 }
 if (int_val & 0x02) {
	while (RXRDYA () ) {
		buffer0[bufferin0] = read (RHRA);
		bufferin0++;
		if (bufferin0 == SIZE) bufferin0 = 0;
	}
 }
 if (int_val & 0x10) {
		while ( TXRDYB () ) {
			write (THRB, data1[databuf1]);
			databuf1++;
			if (databuf1 == 64) databuf1 = 0;
		}
 }
 if (int_val & 0x01) {
		 while ( TXRDYA () ) {
			write (THRA, data0[databuf0]);
			databuf0++;
			if (databuf0 == 64) databuf0 = 0;
		}
 }

  outportb (0x20, 0x20);
  write ( IMR, imr);
}


/***********************************************************************
 The baseaddress for the XR88C92/68C92 board is one of
 0x400, 0x4e0, 0x500 or 0x5e0. This function scans these
 addresses to determine where the chip is located.
**************************************************************************/
int chipscan (void)
{
 int i, status1, status2;
 int base [4] = {0x400, 0x4e0, 0x500, 0x5e0};
 for (i = 0; i < 4; i++ )
 {
  baseaddress = base[i];
  write (CRA, 0x20);	// reset RX
  write (CRA, 0x30);	// reset TX
  write (CRA, 0x05);	// enable TX & RX
  status1 = read (SRA) & 0xff; // Get the status: 0x0C indicates UART presence
  write (CRB, 0x20);	// reset RX
  write (CRB, 0x30);	// reset TX
  write (CRB, 0x05);	// enable TX & RX
  status2 = read (SRB) & 0xff; // Get the status: 0x0C indicates UART presence
  if ( (status1 == 0x0c) && (status2 == 0x0c) )
  {
   cout << "\nUART detected at baseaddress = 0x" << hex << baseaddress << "\n";
   return 1;
  }
 }
cout << "\nNo UART Detected in the system\n";
return -1;
}

int init (void)
{
 write (CRA, 0xf0);	//	Disable Power Down Mode (just in case)

 write (CRA, 0x0a);	// Disable RX & TX
 write (CRB, 0x0a);

 write (CRA, 0x20);  // Reset RX
 write (CRB, 0x20);

 write (CRA, 0x30);	// Reset TX
 write (CRB, 0x30);

 write (CRA, 0x40);	// Reset Error Status
 write (CRB, 0x40);

 write (CRA, 0x50);	// Reset Break-change interrupt
 write (CRB, 0x50);

 write (CRA, 0x90);	// Reset -RTS output to HIGH
 write (CRB, 0x90);

 write (CRA, 0xb0);	// Reset MR pointer to MR0
 write (CRB, 0xb0);

 write (MRA, 0x70 | mr0_bit0);	// now accessing MR0
 write (MRB, 0x70);	// Choose Baud Rate table,
								// TX Trig of 1, RX trig of 6

 write (CRA, 0x10);	// reset MR pointer to MR1 (optional)
 write (CRB, 0x10);

 write (MRA, 0x93);	// now accessing MR1
 write (MRB, 0x93);	// 8N1, Multi-drop mode disabled; auto RTS enabled

 write (MRA, (0x07 | ctsa) );
 write (MRB, (0x07 | ctsb)  );

 write (ACR, acr);	// Set ACR bit-7 = 1 when you want additional
						// baud rates not available otherwise.

 write (CSRA, csra);	// User selected Baud Rate
 write (CSRB, csrb);

 write (CRA, 0x05);	// Enable RX & TX
 write (CRB, 0x05);

// Enable RTS once to start auto RTS flow control
  write (CRA, 0x80);
  write (CRB, 0x80);

 if ( (read (SRA) != 0x0c) || (read (SRB) != 0x0c) )
 {
  cout << "\nInitialization Error";
  return -1;
 }

 return 1;
}


 int main (void)
{
START:
 int char0, char1, i;
 long rxda = 0;
 long rxdb = 0;
 char CR = 0x0d;
 char x92, channel, dummy;
 clrscr();
 if (  chipscan() == -1 ) return -1;
 acr = 0;
 mr0_bit0 = 0;
 csra = 0;
 csrb = 0;
 baud = 0;

 for (i = 0; i < 64; i++) {
  if (i < 62)  {
		data0[i] = i + 0x30;
		data1[i] = i + 0x30;
  } else {
	data0[62] = 0x0a;
	data0[63] = 0x0d;
	data1[62] = 0x0a;
	data1[63] = 0x0d;
  }
 }
 cout << "\nUART Initialization:\n";
 cout << "Press '1' for XR88C92/68C92 and '2' for XR88C192/68C192\n";
 while (!kbhit() );
 x92 = getch();
 if ( x92 == '2') fifo = 16;
 else fifo = 8;
 clrscr();
 cout << "\nUART Initialization:\n";
 cout << "Choose Baud Rate from the following choices:\n\
\t1) 110 bps\n\
\t2) 300 bps\n\
\t3) 600 bps\n\
\t4) 1200 bps\n\
\t5) 2400 bps\n\
\t6) 4800 bps\n\
\t7) 9600 bps [default]\n\
\t8) 19.2 kbps\n\
\t9) 38.4 kbps\n\
\tA) 57.6 kbps\n\
\tB) 115.2 kbps\n\
\tC) 230.4 kbps\n";
	while (! kbhit() );
	baud = getch();
	cout << baud << "\n";
	switch (baud)
	{
	 case '1'	:	csra = 0x11;
					csrb = 0x11;
						break;
	 case '2'	:	csra = 0x44;
					csrb = 0x44;
						break;
	 case '3'	:	csra = 0x55;
					csrb = 0x55;
						break;
	 case '4'	:	csra = 0x66;
					csrb = 0x66;
						break;
	 case '5'	:	csra = 0x88;
					csrb = 0x88;
						break;
	 case '6'	:	csra = 0x99;
					csra = 0x99;
						break;
	 case '7'	:	csra = 0xbb;
					csrb = 0xbb;
						break;
	 case '8'	:	acr = 0x80;
					csra = 0xcc;
					csrb = 0xcc;
						break;
	 case '9'	:	csra = 0xcc;
					csrb = 0xcc;
						break;
	 case 'a'	:	case 'A'	:
						mr0_bit0 = 1;
						csra = 0xbb;
						csrb = 0xbb;
						break;
	 case 'b'	:	case 'B'	:
						mr0_bit0 = 1;
						acr = 0x80;
						csra = 0xcc;
						csrb = 0xcc;
						break;
	 case 'c'	:	case 'C'	:
						mr0_bit0 = 1;
						csra = 0xcc;
						csrb = 0xcc;
						break;
	 default		:	csra = 0xbb;
						csrb = 0xbb;
						break;
	}
 clrscr();

 cout << "\nThe UART channels are now set up to receive data.";
 cout << "\nPress 'A' or 'B' to transmit test pattern on channel A or B";
 cout << "\nPress 'C' to transmit test pattern on both channels";
 cout << "\nPress any other key to stay in receive mode only\n";
 while (! kbhit() );
 channel = getch();
	switch (channel) {
	 case 'a' : case 'A':
				imr = (imr | 0x01);
				ctsa = 0x10;
				break;
	 case 'b' : case 'B':
				imr = (imr | 0x10);
				ctsb = 0x10;
				break;
	 case 'c' : case 'C':
				imr = (imr | 0x11);
				ctsa = 0x10;
				ctsb = 0x10;
				break;
	 default : break;
	}

 if (! init() ) return -1;

  bufferin0 = 0;
  bufferout0 = 0;
  bufferin1 = 0;
  bufferout1 = 0;

  buffer0 = new char[SIZE];
  buffer1 = new char[SIZE];

  oldportisr = getvect(intvect);		//store the current isr
  setvect(intvect, myportisr);		//myportisr is a pointer to an interrupt function

 /* enable IRQ5 in the PIC & enable UART interrupts */
	outportb (0x21, ( inportb(0x21) & 0xdf) );
	write (IMR, imr);

	clrscr();
	cout << "XR88C92/68C92 UART TEST IN PROGRESS";
	cout << "\nAt any time, press any key to exit test program";
	cout << "\n\nNumber of packets received by";
	cout << "\nChannel A\tChannel B\n";
	while ( 1 ) {
		if (bufferin1 != bufferout1) {
			char1 = ( (int)buffer1[bufferout1]) & 0xff ;
			bufferout1++;
			if (bufferout1 == SIZE) bufferout1 = 0;
			rxdb++;
			cout << "\t\t" << dec << (rxdb / 64) << CR;
			if (kbhit() ) goto END;
		}
		if (bufferin0 != bufferout0) {
			char0 = ( (int)buffer0[bufferout0]) & 0xff ;
			bufferout0++;
			if (bufferout0 == SIZE) bufferout0 = 0;
			rxda++;
			cout << dec << (rxda / 64) << CR;
			if (kbhit() ) goto END;
		}
		if (kbhit() ) goto END;
	}
END :
   write (IMR, 0x00);
   outportb (0x21, ( inportb (0x21) | 0x20 ) );
   setvect (intvect, oldportisr);
   cout << "\nProgram Exited\n";
   delete[] buffer0;
   delete[] buffer1;
   write (CRA, 0x20);
   write (CRA, 0x30);
   write (CRA, 0x40);
   write (CRA, 0x50);
   write (CRB, 0x20);
   write (CRB, 0x30);
   write (CRB, 0x40);
   write (CRB, 0x50);
   write (CRA, 0xd0);
   return 1;
}







